import bpy
import bmesh
from mathutils import Matrix
from numpy import dot, arccos, pi, cos, cross
from numpy.linalg import norm
from itertools import zip_longest


class MatchPipeWidth(bpy.types.Operator):
    """Creates a Mesh from a pipe with corrected scaling of the edge loops for a uniform width of the pipe's segments."""
    bl_idname = "object.matchpipewidth"
    bl_label = ""

    def execute(self, context):

        curve = context.object
        spline = curve.data.splines.active
        points = spline.points

        if len(points) >= 3:
            radius_adjustments = [None]
            normals = [None]

            # Calculate Radius Adjustments and Perpendicular Axes for Each Point
            for i in range(1, len(points) - 1):
                a = points[i - 1]
                b = points[i]
                c = points[i + 1]

                ba = a.co.to_3d() - b.co.to_3d()
                bc = c.co.to_3d() - b.co.to_3d()

                alpha = angle(ba, bc)
                beta = pi - alpha

                radius_adjustments.insert(i, abs(1 / cos(beta / 2)))
                normals.insert(i, cross(ba, bc) / norm(cross(ba, bc)))

            ring_size = 4 + 2 * curve.data.bevel_resolution

            # Convert to Mesh
            bpy.ops.object.convert(target='MESH')

            mesh = curve.data

            # Get Bmesh
            bpy.ops.object.mode_set(mode='EDIT')

            if mesh.is_editmode:
                bm = bmesh.from_edit_mesh(mesh)
            else:
                bm = bmesh.new()
                bm.from_mesh(mesh)

            # Group Edge Loops (Assumes that vertices are created ring after ring)
            rings = list(grouper(bm.verts, ring_size))

            for i in range(1, len(rings) - 1):
                # Select Vertices
                for vertex in rings[i]:
                    vertex.select = True

                # Create Rotation Matrices
                rotation_axis = cross(normals[i], (0, 0, 1))
                rotation_angle = angle(normals[i], (0, 0, 1))
                rotation_matrix = Matrix.Rotation(rotation_angle, 4, rotation_axis)
                backrotation_matrix = Matrix.Rotation(-rotation_angle, 4, rotation_axis)

                # Rotate, Scale, Rotate Back
                bmesh.ops.rotate(bm, cent=(0, 0, 0), matrix=rotation_matrix, verts=rings[i])
                bpy.ops.transform.resize(value=(radius_adjustments[i], radius_adjustments[i], 1), constraint_axis=(True, True, False))
                bmesh.ops.rotate(bm, cent=(0, 0, 0), matrix=backrotation_matrix, verts=rings[i])

                # Deselect Vertices
                for vertex in rings[i]:
                    vertex.select = False
        else:
            bpy.ops.object.convert(target='MESH')

        return {'FINISHED'}


def angle(v1, v2):
    angle = arccos(dot(v1, v2) / (norm(v1) * norm(v2)))
    if (angle <= 2 * pi):
        return angle
    else:
        return 2 * pi - angle


def grouper(iterable, n, fillvalue=None):
    args = [iter(iterable)] * n
    return zip_longest(*args, fillvalue=fillvalue)
