import blf
import bpy
from bpy.props import IntProperty, FloatProperty


def draw_pipe_options(self, context):
    if context.object.type == 'CURVE':
        segments = context.object.data.bevel_resolution
        pipe_width = context.object.data.bevel_depth
        width = context.area.width
        font_id = 0
        uif = context.preferences.view.ui_scale
        s = 16 * uif
        y = 60 * uif
        blf.color(font_id, 1, 1, 1, 1)
        blf.position(font_id, width / 2 - 100, y + s * 5, 0)
        blf.size(font_id, s)
        blf.draw(font_id, "Segments: ")

        blf.position(font_id, width / 2 - 100, y + s * 3, 0)
        blf.draw(font_id, "Width: ")

        blf.position(font_id, width / 2 - 100, y + s, 0)
        blf.draw(font_id, "Fill Caps: ")

        blf.color(font_id, 253, 253, 0, 1)
        blf.position(font_id, width / 2 + 35, y + s * 5, 0)
        blf.size(font_id, s)
        blf.draw(font_id, str(4 + segments * 2))

        blf.position(font_id, width / 2 - 14, y + s * 3, 0)
        blf.draw(font_id, str("%.3f" % pipe_width))

        blf.position(font_id, width / 2 + 16, y + s, 0)
        if context.scene.quick_pipe_fill_caps is True:
            blf.draw(font_id, "ON")
        else:
            blf.draw(font_id, "OFF")


def set_status_text(context, text_list, spacing=5, mb="\u25cf", kb="\u25a0"):
    if text_list is not None:
        separator = "\u2003" * spacing
        message = ""
        for line in text_list:
            if "WHEEL" in line or "MB" in line:
                message += mb + " " + line + separator
            else:
                message += kb + " " + line + separator
        context.workspace.status_text_set(message)
    else:
        context.workspace.status_text_set(None)


class jmPipeTool(bpy.types.Operator):
    bl_idname = "object.quickpipe"
    bl_label = "Quick Pipe"
    bl_options = {'REGISTER', 'UNDO'}

    first_mouse_x: IntProperty()
    first_value: FloatProperty()

    @classmethod
    def poll(cls, context):
        return context.object and context.object.type == "MESH" and context.mode == "EDIT_MESH"

    def separate_curves(self, context):
        # Agnieszka Pas "Separate Curve Object"
        # https://github.com/agapas/separate-curves/blob/master/separate_curves.py
        bpy.ops.object.mode_set(mode='EDIT')
        bpy.ops.curve.select_all(action='DESELECT')
        splines = context.active_object.data.splines
        while len(splines) > 1:
            splinee = splines[0]
            if splinee.points:
                splinee.points[0].select = True
            bpy.ops.curve.select_linked()
            bpy.ops.curve.separate()
        bpy.ops.object.mode_set(mode='OBJECT')

    def fill_caps(self):
        bpy.ops.object.mode_set(mode='EDIT')
        bpy.ops.mesh.select_all(action='SELECT')
        bpy.ops.mesh.region_to_loop()
        bpy.ops.mesh.fill_holes(sides=100)
        bpy.ops.mesh.poke()
        bpy.ops.mesh.select_all(action='DESELECT')
        bpy.ops.object.mode_set(mode='OBJECT')

    def modal(self, context, event):
        addon_prefs = context.preferences.addons["quick_pipe"].preferences

        if event.type in {'RIGHTMOUSE', 'LEFTMOUSE', 'SPACE'}:
            context.object.show_wire = False
            self.separate_curves(context)
            if addon_prefs.convert_to_mesh is True:
                bpy.ops.object.convert(target='MESH')
            bpy.ops.object.mode_set(mode='OBJECT')
            if context.scene.quick_pipe_fill_caps is True:
                bpy.ops.object.convert(target='MESH')
                self.fill_caps()

            context.area.header_text_set(None)
            context.workspace.status_text_set(None)
            bpy.types.SpaceView3D.draw_handler_remove(self._handle, 'WINDOW')
            return {'FINISHED'}

        if event.type == 'Q':
            context.object.show_wire = False
            self.separate_curves(context)

            # Match pipe width for each curve
            pipe_obj = [obj for obj in context.selected_objects]

            for obj in pipe_obj:
                bpy.ops.object.mode_set(mode='OBJECT')
                bpy.ops.object.select_all(action='DESELECT')
                obj.select_set(state=True)
                context.view_layer.objects.active = obj
                if context.object.type == 'CURVE':
                    bpy.ops.object.matchpipewidth()

            bpy.ops.object.mode_set(mode='OBJECT')
            # Select all pipes
            for obj in pipe_obj:
                obj.select_set(state=True)

            if context.scene.quick_pipe_fill_caps is True:
                self.fill_caps()

            context.area.header_text_set(None)
            context.workspace.status_text_set(None)
            bpy.types.SpaceView3D.draw_handler_remove(self._handle, 'WINDOW')
            return {'FINISHED'}

        if event.type == 'W' and event.value == "PRESS":
            context.object.show_wire = not context.object.show_wire

        if event.type == 'S' and event.value == "PRESS":
            if context.object.data.splines[0].use_smooth is True:
                bpy.ops.object.shade_flat()
            else:
                bpy.ops.object.shade_smooth()

        if event.type == 'F' and event.value == "PRESS":
            context.scene.quick_pipe_fill_caps = not context.scene.quick_pipe_fill_caps
            context.area.tag_redraw()

        if event.type == 'C':
            # Convert to mesh
            context.object.show_wire = False
            self.separate_curves(context)
            bpy.ops.object.convert(target='MESH')
            context.area.header_text_set(None)
            context.workspace.status_text_set(None)
            bpy.types.SpaceView3D.draw_handler_remove(self._handle, 'WINDOW')

            if context.scene.quick_pipe_fill_caps is True:
                self.fill_caps()
            return {'FINISHED'}

        if event.type == 'MIDDLEMOUSE':
            bpy.ops.view3d.rotate("INVOKE_DEFAULT")
            return {'PASS_THROUGH'}

        if event.type == 'MOUSEMOVE':
            delta = (self.first_mouse_x - event.mouse_x)
            if event.ctrl:
                delta *= 0.01
            if event.shift:
                delta *= 0.1
            context.object.data.bevel_depth = abs((self.first_value + delta) * 0.002)

        if event.type == 'D' and event.value == "PRESS":
            context.object.data.bevel_resolution += 1

        if event.type == 'WHEELUPMOUSE':
            context.object.data.bevel_resolution += 1

        if event.type == 'A' and event.value == "PRESS":
            if context.object.data.bevel_resolution > 0:
                context.object.data.bevel_resolution -= 1

        if event.type == 'WHEELDOWNMOUSE':
            if context.object.data.bevel_resolution > 0:
                context.object.data.bevel_resolution -= 1

        segments = context.object.data.bevel_resolution
        pipe_width = context.object.data.bevel_depth
        return {'RUNNING_MODAL'}

    def invoke(self, context, event):
        context.object.update_from_editmode()
        sel_edges = [e for e in context.object.data.edges if e.select]
        if not sel_edges:
            self.report({"INFO"}, "No edges selected")
            return {"CANCELLED"}

        addon_prefs = context.preferences.addons["quick_pipe"].preferences
        context.scene.quick_pipe_fill_caps = False
        args = (self, context)
        self._handle = bpy.types.SpaceView3D.draw_handler_add(draw_pipe_options, args, 'WINDOW', 'POST_PIXEL')
        self.first_mouse_x = event.mouse_x

        source_objects = [obj for obj in context.selected_objects]
        # Separate selected items
        bpy.ops.mesh.duplicate()
        bpy.ops.mesh.separate(type='SELECTED')
        bpy.ops.object.mode_set(mode='OBJECT')

        all_objects = [obj for obj in context.selected_objects]
        # Find separated objects
        separated_objects = [obj for obj in all_objects if obj not in source_objects]

        bpy.ops.object.select_all(action='DESELECT')
        # Select separated objects
        for obj in separated_objects:
            obj.select_set(state=True)

        separated_objects[0].name = "Pipe"
        context.view_layer.objects.active = separated_objects[0]
        if len(source_objects) > 1:
            bpy.ops.object.join()

        if addon_prefs.display_wire is True:
            context.object.show_wire = True

        bpy.ops.object.convert(target='CURVE')

        if addon_prefs.enable_smooth_shading is True:
            bpy.ops.object.shade_smooth()

        pipe = context.view_layer.objects.active
        pipe.data.fill_mode = 'FULL'
        pipe.data.bevel_resolution = 1
        pipe.data.bevel_depth = 0.1

        self.first_value = pipe.data.bevel_depth
        context.window_manager.modal_handler_add(self)
        text = [
            "MMB: Rotate View", "CTRL: 0.01 precision", "SHIFT: 0.1 precision",
            "RMB/LMB/Space: Apply", "MWHEEL/A,D: Segments",
            "Q: Match Width", "W: Wireframe", "S: Smooth shading", "F: Fill caps(toggle)", "C: Convert to mesh"
        ]
        set_status_text(context, text_list=text)
        return {'RUNNING_MODAL'}
